;--------------------------------------------------------------------------
;
;	MFM Writer
;
;	1. Write Sector
;
;	When writing a sector we need to write the complete data segment
;
;	+--------+----+----+----------------------------+-----+---------+
;	|preamble|sync|DAM |             Data           | CRC |postamble|
;	|  12 x  | 3x | 1x |            512 x           | 2x  |   40 x  |
;	|  0x00  |0xA1|0xFB|            user            | CRC |   0x4E  |
;	+--------+----+----+----------------------------+-----+---------+
;
;	First we need to find the correct sector using the "readsection"
;	routine. When we have found the correct ID record we need to 
;	start writing after the GAP that follows the ID record. Thats
;	why the "readsection" routine always initiates a timer when it
;	has itself sync'd to the preamble and the sync marks. We need to
;	start writing when the timer overflows. 
;
	.macro	w6cycles
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	.endmacro
	
	.macro	wcycles
	.set cycles = @0
	.if ((cycles < 8) || (cycles > 1000))
	.error "Count must be between 8 and 1000)"
	.endif
	push	count
	ldi		count, (cycles/4)-1
l1:
	nop
	dec		count
	brne	l1
	pop		count
	.if (cycles & 1)
	nop
	.endif
	.if (cycles & 2)
	rjmp	PC+1
	.endif
	.endmacro
;
;	xxxx<yyyy^W		Write Buffer at yyyy to Sector xxxx
;
;		a4	= Sector Address
;		a1	= Buffer Address with data
;
;	Note that the MFMReader uses all the memory from 0x1000 to 0x3fff
;	but there is still enough room below 0x1000 to place the buffer
;	we want to write. So if you want to crate patterns and data to
;	be written use a buffer below 0x1000 (the MFMRead uses only very
;	little low memory that starts at 0x100)
;

writepattern:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	lds		temp, a4l			; get sector requested
	tst		temp
	brne	writepattern005
	mov		temp, one
writepattern005:
	cpi		temp, numsect
	brlo	writepattern010
	ldi		temp, numsect
writepattern010:
	sts		writesector, temp
;
;	Prepare the sectorbuffer we want to write. The buffer includes
;	the Data Mark (0xFB), data, CRC and the GAP (22*0x4E). Also
;	generate the CRC.
;
	lds		xl, a1l
	lds		xh, a1h				; get memory buffer to write
	
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)
	mov		crcl, ff			; Initiate CRC
	mov		crch, ff
	ldi		temp, 0xA1			; Note CRC calculation inlcudes
	updcrc	temp				; the sync marks and the data mark
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xFB
	st		Y+, temp			; Save data mark in buffer
	updcrc	temp
	clr		count				; Works only for sector sizes of 512bytes
								; but without someone telling us we don't
								; know, we could read an arbitrary ID record
								; to retrieve the sector length but that could
								; affect performance, on the other hand we could
								; also remember the sector length form any ID
								; record and keep that as reference, it is very
								; unlikely that a floppy is formatted with
								; different sector sizes, although it could.
								; CRC generation on the fly is not working
								; the timing is too tight.

writepattern020:
;
;	Each Loop deals with 2 bytes
;
	ld		temp, X+			; copy memory buffer to sector buffer
	st		Y+, temp
	updcrc	temp				; and update CRC at the same time
	ld		temp, X+
	st		Y+, temp
	updcrc	temp
	dec		count
	brne	writepattern020
	

	st		Y+, crch			; Add CRC
	st		Y+, crcl
	ldi		temp, 0x4E			; Add GAP
	ldi		count, 48
writepattern040:
	st		Y+, temp
	dec		count
	brne	writepattern040

	rcall	write_nextidsegment
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret
	
write_nextidsegment:
	ldi		xl, low(7)				; Bytes to collect
	ldi		xh, high(7)
	ldi		yl, low(idsegment)
	ldi		yh, high(idsegment)
	clr		miss
	rcall	readsection
	brcc	write_readok
	ldi		char, '?'
	call	serout
	ret
	
;
;	And do now the "house-keeping"
;
write_readok:
	mov		crcl, ff
	mov		crch, ff
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	lds		temp, idsegment
	cpi		temp, 0xFE
	brne	write_nextidsegment		; It's not an idsegment
	updcrc	temp
	lds		temp, idsegment+1		; Track
	updcrc	temp
	lds		temp, idsegment+2		; Side
	updcrc	temp
	lds		temp, idsegment+3		; Sector
	updcrc	temp
	lds		temp, idsegment+4		; Length
	updcrc	temp
	lds		temp, idsegment+5
	cp		crch, temp
	lds		temp, idsegment+6
	cpc		crcl, temp
	breq	write_crcok
	ldi		char, '!'
	call	serout
	ret
write_crcok:
	lds		temp, idsegment+3
	lds		r0, writesector
	cp		temp, r0
	breq	write_waitgap
	rjmp	write_nextidsegment
	
;
;	The readsection always starts timer 1 when it has found a sync mark.
;	We only need to wait for the overflow flag to know when we need to
;	start writing the sector
;

write_waitgap:
write_waitgap010:
	sbis	TIFR1, TOV1
	rjmp	write_waitgap010
	cbi		PORTC, WG	
;
;
;	95 short intervals ......SSSSSSSSSSSSSS
;
	ldi		count, 12*8
wp0010:
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	19
	dec		count
	brne	wp0010
	nop
;
;	followed by three Sync Marks MLMLMSLMLMSLMLM
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	22				; S
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	22				; S
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38-6			; M-shortened to compensate initialisation
;
;	We use the following defintions
;
;	T-bit		Previous Clock Type '1'=data, '0'=clock
;	Char		The current byte
;	Count		Number of bits left in char
;	Y			Pointer to the buffer
;	X			Number of bytes to transmit (must be positive number)
;
;	Initiate the state
;
	set							; Sync Mark ends with a data pulse
	clr		count				; We have no bits in char
	ldi		xl, low(1+512+2+38)
	ldi		xh, high(1+512+2+38)
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)

;	We are now here and need to create the pulse to finishe the
;	last M interval of the third SYNC MARK
;			------------->
;	Data	1 0 1 0 0 0 0 1| 1
;	Clock	 0 0 0 1 1 1 0 |0
;	MFM		100010010001001|01
;
;
;	Logic to generate MFM is very simple. We need only to remember
;	one state, that is if the previous pulse was a clock or a data
;	pulse.
;
;	get_next_bit()
;	if (previous_pulse == clock_pulse)
;	{
;		if (next_data_bit == '0')
;		{
;			prev next
;			---- ---- 
;	Data:	   0|   0|
;	Clock:	 1  | 1  |
;	MFM:	0100 0100 
;
;			interval=2usec
;			previous_pulse=clock_pulse
;		}
;		else
;		{
;			prev next
;			---- ---- 
;	Data:	   0|   1|
;	Clock:	 1  | 0  |
;	MFM:	0100 0001 
;
;			interval=3usec
;			previous_pulse=data_pulse
;		}
;	}
;	else
;	{
;		if (next_data_bit == '0')
;		{
;			get_next_bit()
;			if (next_data_bit == '0')
;			{
;			prev next
;			---- ---- ---- 
;	Data:	   1|   0|   0|
;	Clock:	 0  | 0  | 1  |
;	MFM:	0001 0000 0100 
;
;				interval=3usec
;				prevous_pulse=clock_pulse
;			}
;			else
;			{
;
;			prev next
;
;			---- ---- ---- 
;	Data:	   1|   0|   1|
;	Clock:	 0  | 0  | 0  |
;	MFM:	0001 0000 0001 
;
;			interval=4usec
;			prevous_pulse=data_pulse
;		}
;		else
;		{
;			prev next
;			---- ---- 
;	Data:	   1|   1|
;	Clock:	 0  | 0  |
;	MFM:	0001 0001 
;
;			interval=2usec
;			previous_pulse=data_pulse
;		}
;	}
;
;	We first generate a pulse using first a cbi some nop and a sbi
;	In case we have a clock of 16MHz and we want to create pulses
;	of 500ns this looks like the following.
;
w0000:
	cbi		PORTC, WD
	rjmp	PC+1				; rjmp PC+1 uses one instruction word
	rjmp	PC+1				; and 2 cycles, its 2 nops in one instruction
	rjmp	PC+1
	sbi		PORTC, WD
;
;	Each pulse is followed by at least 1500nsec silence. This corresponds
;	to 24 cycles. However the pulse generation already took 2 cycles so
;	we have 22 cycles left.
;									cycles
w0010:
	dec		count				;	1	More bits to go
	brmi	w0020				;	2	Need a byte

	nop							;		Compensate branch not taken
	rjmp	PC+1				;		Instead of getting a new byte
	rjmp	PC+1				;	 	we need to waste time
	rjmp	w0030				;
;
;	Getting a new byte
;	-	if we reached the end of the buffer then we are done
;	-	else set the number of bits we need to process to 7
;		because we are going to immediately process a bit
;	-	get the new byte
;

w0020:
	sbiw	X, 1				;	2	More bytes to go
	brmi	w0090				; 	1	No we are done
	ldi		count, 7			;	1	We will have 7 bits left
	ld		char, Y+			;	2	Get next char from buffer
w0030:
	lsl		char				;	1	Get next bit to the Carry bit
								;  ==
								;  10	cycles
;	We now have 4 possibilities
;
;	T=C=0
;	T=0, C=1
;	T=1, C=0
;	T=C=1
;
;	Some info about T-bit processing:
;	If interval is an even number of usec then T does not change if it is
;	odd it changes, in other words it changes whenever the interval is 3usec.
;
	brts	w0050				;		decision branch 1 or 2 cycle
	brcs	w0040				;		decision branch 1 or 2 cycle
;
;	T=C=0
;
;
								;	2	cycles from two branches not taken
;
;	Required Interval 2usec				 32 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 2	cycles
;	Cycles left							 10	cycles
;
;
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
								;  ==
								;  10	cycles
;
;	Exit hook so we can always use branches
;
w0090:
	sbi		PORTC, WG
	ret
;	rjmp	wexit				; HOOK


;
w0040:
;
;	TC=0, C=1
;
								;	3	cycles from first branch not taken
								;		and second branch taken
;
;	Required Interval 3usec				 48 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Cycles left							 25	cycles
;
	set							;	1
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
								;  ==
								;  25	cycles
;
w0050:
	brcs	w0080				;
;
;	T=1, C=0
;
								;	3	cycles from first branch taken
								;		and second branch not taken
;
;	We need another bit
;
	dec		count				;	1	More bits to go
	brmi	w0060				;	2	Need a byte

	nop							;		compensate branch not taken
	rjmp	PC+1				;		Instead of getting a new byte
	rjmp	PC+1				;		we need to waste time
	rjmp	w0070


w0060:
	sbiw	X, 1				;	2	More bytes to go
	brmi	w0090				; 	1	No we are done
	ldi		count, 7			;	1	We will have 7 bits left
	ld		char, Y+			;	2	Get next char from buffer

w0070:
	lsl		char				;	1
								;  ==
								;  10	cycles
;
	brcc	w0075				;
;
;	Next bit is '1'
;
;	Required Interval 4usec				 64 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Check for new character				-10 cycles
;	Decision Branch						- 1	cycle
;	Cycles left							 30	cycles
;
;	
;
;
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
								;  ==
								;  30
;
;	Next bit is '0'
;
;	Required Interval 3usec				 48 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Check for new character				-10 cycles
;	Decision Branch						- 2	cycle
;	Cycles left							 13	cycles
;
w0075:
	clt							;	1
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
;
w0080:
;
;	T=C=1
;
								;	4	cycles from both branch taken
;
;	Required Interval 2usec				 32 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 4	cycles
;	Cycles left							  8	cycles
;
;	So far we have spent 14 cycles (10 cycles from the buffer check and 2 x 2
;	cycles from branches taken). Previous pulse was a data pulse and next
;	bit is '1' so the interval is usec, so we need to waste 8 cycles
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
































